/*
=============================================================================
Module Information
------------------
Name:			obj_soundemit.cpp
Author:			Rich Whitehouse
Description:	server logic object: sound emitter

Object Arguments
sound:			Name of the sound to emit.
delay:			Interval in milliseconds between emission.
volume:			0.0 to 1.0 volume scale of the sound.
onetime:		Remove after emitting the sound.
startoff:		Starts off, must be activated to begin emission.
=============================================================================
*/

#include "main.h"

#define SEFL_ONETIME		(1<<1)
#define SEFL_ENABLED		(1<<2)

//frame function
void ObjSoundEmit_Think(gameObject_t *obj, float timeMod)
{
	if (!(obj->generalFlag & SEFL_ENABLED))
	{
		obj->thinkTime = g_glb.gameTime;
		return;
	}

	int objAttach = -1;
	if (obj->target && obj->target->inuse)
	{
		objAttach = obj->target->net.index;
	}
	ObjSound_CreateFromIndex(obj->net.pos, obj->net.strIndex, obj->net.modelScale[0], objAttach);
	obj->thinkTime = g_glb.gameTime+obj->debounce;
	if (obj->generalFlag & SEFL_ONETIME)
	{ //one-time
		obj->think = ObjGeneral_RemoveThink;
	}
}

//activate
void ObjSoundEmit_Activate(gameObject_t *obj, gameObject_t *activator)
{
	obj->generalFlag ^= SEFL_ENABLED;
}

//spawn
void ObjSoundEmit_Spawn(gameObject_t *obj, BYTE *b, const objArgs_t *args, int numArgs)
{
	obj->localFlags |= LFL_NONET;
	obj->inuse |= INUSE_NONET; //don't need emitters sent to the client ever.

	int oneTime = 0;
	int delay = 50;
	obj->net.modelScale[0] = 1.0f; //default volume
	const char *sound = NULL;
	obj->generalFlag = SEFL_ENABLED;
	if (obj->spawnArgs && obj->numSpawnArgs > 0)
	{
		int i = 0;
		while (i < obj->numSpawnArgs)
		{
			const objArgs_t *arg = obj->spawnArgs+i;
			if (!_stricmp(arg->key, "onetime"))
			{
				oneTime = atoi(arg->val);
			}
			else if (!_stricmp(arg->key, "sound"))
			{
				sound = arg->val;
			}
			else if (!_stricmp(arg->key, "delay"))
			{
				delay = atoi(arg->val);
			}
			else if (!_stricmp(arg->key, "volume"))
			{
				obj->net.modelScale[0] = (float)atof(arg->val);
			}
			else if (!_stricmp(arg->key, "startoff"))
			{
				obj->generalFlag &= ~SEFL_ENABLED;
			}
			i++;
		}
	}

	if (!sound)
	{
		obj->think = ObjGeneral_RemoveThink;
		obj->thinkTime = g_glb.gameTime+50;
		return;
	}

	obj->debounce = (serverTime_t)delay;

	obj->think = ObjSoundEmit_Think;
	if (oneTime)
	{
		obj->generalFlag |= SEFL_ONETIME;
		obj->thinkTime = g_glb.gameTime+obj->debounce;
	}
	else
	{
		obj->thinkTime = g_glb.gameTime+50;
	}

	obj->activate = ObjSoundEmit_Activate;

	char str[64];
    sprintf(str, "$%s", sound);
	obj->net.strIndex = g_sharedFn->Common_ServerString(str);
}
